/*
 * dcfp - definition and configuration file parser
 * Copyright (c) 2002 Mark Smulders <Mark@PIRnet.nl>             
 *
 * This program is free software; you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation; either version 2, or (at your option)   
 * any later version.                                                    
 *                                                                       
 * This program is distributed in the hope that it will be useful,       
 * but WITHOUT ANY WARRANTY; without even the implied warranty of        
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         
 * GNU General Public License for more details.                          
 *                                                                       
 * You should have received a copy of the GNU General Public License     
 * along with this program; if not, write to the                         
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,          
 * Boston, MA 02111-1307, USA.                                           
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dcfp.h"

//#include <dmalloc.h>

/*
   For an explanation of the dcfp data structure, a conf file example and an
   example of typical use, please see the header file (dcfp.h)
*/

/*********************** DCFP PRIVATE FUNCTION PROTOTYPES ************************/

static char *dcfp_get_key_from_line(char *line);
static char *dcfp_get_value_from_line(char *line);
static char *dcfp_parse_section_name(char *readline);
static void dcfp_parse_section(dcfp_section_struct *sectptr, FILE *fp, char *line, int *line_nr);
static void dcfp_parse_var(dcfp_var_struct *varptr, char *line);
static char *dcfp_rm_first_spaces(const char *ptr);
static char *dcfp_get_single_line_without_first_spaces(FILE *fp, int *line_nr);
static char *dcfp_dyn_fgets(FILE *fp);
static void dcfp_destroy_var_tree(dcfp_var_struct *vars);
static void dcfp_destroy_section(dcfp_section_struct *sect);
static int is_int (char *str);
static void *dcfp_get_var_from_varlist(dcfp_var_struct *varlist, const char *varname);
static dcfp_section_struct *dcfp_get_section_from_sectionlist_by_name(dcfp_section_struct *sectionlist, const char *name);

/*********************** DCFP INTERFACE (PUBLIC) FUNCTIONS ***********************/

/*
  dcfp_parse_file()
  parse the contents of a file into the dynamic struct 
*/
const dcfp_data_struct *dcfp_parse_file(const char *filename) {
  dcfp_data_struct *data = NULL;
  dcfp_var_struct *varptr=NULL;
  dcfp_section_struct *sectptr=NULL;
  char *line=NULL;
  int line_nr = 0, firstvar=1, firstsect=1;
  FILE *fp=NULL;
  
  /* init data struct */
  data = (dcfp_data_struct *)malloc(sizeof(dcfp_data_struct));
  data->globalvars = NULL;
  data->sections = NULL;
  
  /* open the file */
  if ((fp=fopen(filename, "r")) == NULL) {
	// TODO: INSERT ERROR HANDLER
	free(data);
	return NULL;
  }

  /* parse the contents */
  while (!feof(fp)) {
	if ((line = dcfp_get_single_line_without_first_spaces(fp, &line_nr))==NULL) { break; }
	if ( *line == '[' ) { //global section start found
	  if ((char *)strstr(line,"[END")!=NULL) {
		//TODO: INSERT ERROR HANDLER (instead of code below)
		printf("*** Error found in file %s (line %d)\n*** Make sure there's only one [END] for each [SECTION]\n*** Aborting...\n",filename,line_nr);
		dcfp_destroy_data_struct(data);
		fclose(fp);
		return NULL;
	  }
      if (firstsect) {
	    data->sections = (dcfp_section_struct *)malloc(sizeof(dcfp_section_struct));
	    sectptr = data->sections;
	    firstsect = 0;
	  } else {
	    sectptr->next = (dcfp_section_struct *)malloc(sizeof(dcfp_section_struct));
	    sectptr = sectptr->next;
	  }
      dcfp_parse_section(sectptr, fp, line, &line_nr);
	} else { //global var found
	  if (firstvar) {
		data->globalvars = (dcfp_var_struct *)malloc(sizeof(dcfp_var_struct));
		varptr = data->globalvars;
		firstvar = 0;
	  } else {
        varptr->next = (dcfp_var_struct *)malloc(sizeof(dcfp_var_struct));
		varptr = varptr->next;
	  }
	  dcfp_parse_var(varptr, line);
	}
	free(line); line=NULL;
  }

  /* close file and exit */
  fclose(fp);
  return data;
}

/*
  dcfp_destroy_data_struct()
  free the entire dynamic data tree from memory
*/
void dcfp_destroy_data_struct(dcfp_data_struct *data) {
  if (data != NULL ) {
    // free global vars
    dcfp_destroy_var_tree(data->globalvars);
    data->globalvars=NULL;
    // free all sections and subsections
    dcfp_destroy_section(data->sections);
    data->sections=NULL;
    // free the data struct itself
    free(data);
    data=NULL;
  }
}

/*
  dcfp_section_exist()
  return whether or not a global section exists by name
*/
int dcfp_global_section_exist(const dcfp_data_struct *data, const char *sectionname) {
  int sectfound = 0;
  dcfp_section_struct *ptr = NULL;
  if (data==NULL || sectionname==NULL) return sectfound;
  ptr = data->sections;
  while (ptr!=NULL) {
	if (strcasecmp(ptr->name, sectionname)==0) {
	  sectfound = 1;
	  break;
	}
	ptr = ptr->next;
  }
  return sectfound;
}

/*
  dcfp_get_global_section_by_name()
  return the global section called 'name', NULL if not found
*/
const dcfp_section_struct *dcfp_get_global_section_by_name(const dcfp_data_struct *data, const char *name) {
  if (data!=NULL)
    return dcfp_get_section_from_sectionlist_by_name(data->sections, name);
  else
	return NULL;
}

/*
  dcfp_get_subsection_by_name()
  return the subsection from parentsection called 'name', NULL if not found
*/
const dcfp_section_struct *dcfp_get_subsection_by_name(const dcfp_section_struct *parentsection, const char *name) {
  if (parentsection!=NULL)
    return dcfp_get_section_from_sectionlist_by_name(parentsection->sections, name);
  else
	return NULL;
}

/*
  dcfp_get_global_var()
  returns the value (either int* or char*) of the global var by name, NULL if not found
*/
const void *dcfp_get_global_var(const dcfp_data_struct *data, const char *varname) {
  if (data!=NULL)
    return dcfp_get_var_from_varlist(data->globalvars, varname);
  else
	return NULL;
}

/*
  dcfp_get_var_from_section()
  return the value (either int* or char*) of a var from a section by name, NULL if not found
*/
const void *dcfp_get_var_from_section(const dcfp_section_struct *section, const char *varname) {
  if (section!=NULL)
    return dcfp_get_var_from_varlist(section->globalvars, varname);
  else
	return NULL;
}

/*********************** DCFP INTERNAL (PRIVATE) FUNCTIONS ***********************/

/*
  dcfp_get_key_from_line()
  get the key from a line (key = value)
*/
static char *dcfp_get_key_from_line(char *line) {
  char *key=NULL, *ptr=NULL;
  int i=0;
  if (line==NULL) return NULL;
  key = (char *)malloc(sizeof(char));
  *key = '\0';
  ptr = line;
  while (*ptr!=' ' && *ptr!='\t' && *ptr!='=') {
	key = (char *)realloc(key, (i+2)*sizeof(char));
	key[i] = *ptr;
	key[i+1] = '\0';
	i++; ptr++;
  }
  return key;
}

/*
  dcfp_get_value_from_line()
  get the value from a line (key = value)
*/
static char *dcfp_get_value_from_line(char *line) {
  char *value=NULL, *ptr=NULL, *firstptr=NULL;
  int i=0;
  if (line == NULL) return NULL;
  value = (char *)malloc(sizeof(char));
  *value = '\0';
  firstptr = strchr(line, '=');
  firstptr++;
  ptr = dcfp_rm_first_spaces(firstptr);
  if (ptr==NULL)
	return value;
  firstptr = ptr;
  if (*ptr == '"') {
	ptr++;
    while (*ptr!='\n' && *ptr!='\0' && *ptr!='"') {
	  value = (char *)realloc(value, (i+2)*sizeof(char));
	  value[i] = *ptr;
	  value[i+1] = '\0';
	  i++; ptr++;
    }
  } else {
    while (*ptr!='\n' && *ptr!='\0' && *ptr!=' ' && *ptr!='\t') {
	  value = (char *)realloc(value, (i+2)*sizeof(char));
	  value[i] = *ptr;
	  value[i+1] = '\0';
	  i++; ptr++;
    }
  }
  free(firstptr);
  return value;
}

/*
  dcfp_parse_section_name()
  get the section name from a line: [NAME] -> NAME
*/
static char *dcfp_parse_section_name(char *readline) {
  char *secname, *ptr, *endptr;
  int i;
  if (readline == NULL) return NULL;
  secname = (char *)malloc(sizeof(char));
  *secname = '\0';
  ptr = index(readline,'[');
  if ((endptr = rindex(readline,']')) == NULL) {
	//TODO: insert error handler
	endptr = &(readline[strlen(readline)-1]);
  } else
    endptr--;
  for (i=0,ptr++; ptr<=endptr && *ptr!='\0'; ptr++,i++) {
	secname = (char *)realloc(secname, (i+1)*sizeof(char));
    secname[i] = *ptr;
  }
  secname = (char *)realloc(secname, (i+1)*sizeof(char));
  secname[i] = '\0';
  return secname;
}

/*
  dcfp_parse_section()
  parse a section from the file [fp] starting at line [line] into [sectptr]
*/
static void dcfp_parse_section(dcfp_section_struct *sectptr, FILE *fp, char *line, int *line_nr) {
  int firstvar = 1;
  int firstsub = 1;
  dcfp_var_struct *varptr=NULL;
  dcfp_section_struct *subptr=NULL;
  sectptr->next = NULL;
  sectptr->globalvars = NULL;
  sectptr->sections = NULL;
  sectptr->name = dcfp_parse_section_name(line);
//  sectcount++;
  while (!feof(fp)) {
	if ((line = dcfp_get_single_line_without_first_spaces(fp, line_nr))==NULL) return;
	if (*line == '[') {
	  if ( (char *)strstr(line,"[END")!=NULL) { //end section
		free(line);
		break; 
	  } else { //new sub section
		if (firstsub) {
		  sectptr->sections = (dcfp_section_struct *)malloc(sizeof(dcfp_section_struct));
		  subptr = sectptr->sections;
		  firstsub = 0;
		} else {
		  subptr->next = (dcfp_section_struct *)malloc(sizeof(dcfp_section_struct));
		  subptr = subptr->next;
		}
		 // recursively into subsection
	    dcfp_parse_section(subptr,fp,line,line_nr);
	  }
	} else { //var in section
	  if (firstvar) {
	    sectptr->globalvars = (dcfp_var_struct *)malloc(sizeof(dcfp_var_struct));
		varptr = sectptr->globalvars;
		firstvar = 0;
	  } else {
	    varptr->next = (dcfp_var_struct *)malloc(sizeof(dcfp_var_struct));
		varptr = varptr->next;
	  }
      dcfp_parse_var(varptr, line);
	}
	free(line); line=NULL;
  }
}

/*
  dcfp_parse_var()
  parse a key=value pair from line [line] into [varptr]
*/
static void dcfp_parse_var(dcfp_var_struct *varptr, char *line) {
  char *tmpval=NULL;
  
//  varcount++;
  varptr->next=NULL;
  varptr->key = dcfp_get_key_from_line(line);
  tmpval = dcfp_get_value_from_line(line);
  if (*tmpval!='\0' && is_int(tmpval)) {
	strcpy(varptr->type,"int\0");
	varptr->value = (int *)malloc(sizeof(int));
	*(int *)(varptr->value) = atoi(tmpval);
  } else {
	strcpy(varptr->type,"str\0");
	if (*tmpval == '\0') {
	  varptr->value = (char *)malloc(sizeof(char));
	  *((char *)(varptr->value)) = '\0';
	} else {
  	  varptr->value = (char *)malloc((strlen(tmpval)+1)*sizeof(char));
	  strcpy((char *)(varptr->value),tmpval);
	}
  }
  free(tmpval);
}

/*
  dcfp_get_var_from_varlist()
  return the value of a variable from a varlist by name
*/
static void *dcfp_get_var_from_varlist(dcfp_var_struct *varlist, const char *varname) {
  dcfp_var_struct *ptr=NULL;
  if (varlist==NULL || varname==NULL) return NULL;
  ptr = varlist;
  while (ptr!=NULL) {
	if (strcasecmp(ptr->key, varname)==0)
	  return ptr->value;
	ptr=ptr->next;
  }
  return NULL;
}

/*
  dcfp_get_section_from_sectionlist_by_name()
  return the section called 'name' from a sectionlist
*/
static dcfp_section_struct *dcfp_get_section_from_sectionlist_by_name(dcfp_section_struct *sectionlist, const char *name) {
  dcfp_section_struct *ptr=NULL;
  if (sectionlist==NULL || name==NULL) return NULL;
  ptr = sectionlist;
  while (ptr!=NULL) {
	if (strcasecmp(ptr->name, name)==0) {
	  return ptr;
	}
	ptr = ptr->next;
  }
  return NULL;
}

/*
  dcfp_rm_first_spaces()
  produce a new string, but remove lead-off spaces and tabs from original
*/
static char *dcfp_rm_first_spaces(const char *ptr) {
  char *out=NULL, *tmp=NULL;
  if (ptr==NULL || *ptr=='\0') return NULL;
  tmp = (char *)ptr;
  while (*tmp == ' ' || *tmp == '\t')
    tmp++;
  if (*tmp == '\0') return NULL;
  out = malloc((strlen(tmp)+1)*sizeof(char));
  sprintf(out, "%s", tmp);
  return (out);
}

/*
  dcfp_get_single_line_without_first_spaces()
  get a single line without lead-off spaces and tabs from file,
  skip empty lines and those starting with a #
*/
static char *dcfp_get_single_line_without_first_spaces(FILE *fp, int *line_nr) {
  char *readline=NULL, *outline=NULL;
  for (;;) {
	readline=dcfp_dyn_fgets(fp);
	(*line_nr)++;
    if (fp==NULL || readline==NULL)
      return (NULL);
    outline = dcfp_rm_first_spaces(readline);
    free(readline);
	readline=NULL;
    if (outline!=NULL && *outline != '#' && *outline != '\0') {
      return (outline);
    } else {
	  if (outline != NULL) {
		free(outline);
		outline=NULL;
	  }
	}
  }
}

/*
  dcfp_dyn_fgets()
  fgets with dynamic allocated memory
*/
static char *dcfp_dyn_fgets(FILE *fp) {
/*  char *ptr;
  char temp[2];
  int i;

  if (fp == NULL) return NULL;
  ptr = (char *)malloc(sizeof(char));
  if (ptr == NULL) {
    //TODO: INSERT ERROR HANDLER
    return (NULL);
  }
  *ptr = '\0';
  for (i=0 ;; i++) {
    if (fgets(temp, 2, fp) == NULL) {
      if (ferror(fp) != 0 || i == 0) {
        free(ptr);
        return (NULL);
      }
      return (ptr);
    }
    ptr = (char *)realloc(ptr, (strlen(ptr) + strlen(temp) + 1)*sizeof(char));
    if (ptr == NULL) {
      //TODO: INSERT ERROR HANDLER
      return (NULL);
    }
    strcat(ptr, temp);
    if ((char *)strchr(temp, '\n') != NULL) {
      *((char *)strchr(ptr, '\n')) = '\0';
      return (ptr);
	}
  } */
  char buffer[2048], *ptr;
  if (fp==NULL) return NULL;
  /* read line untill we have a good one (not empty) */
  if (fgets(buffer, 2048, fp) == NULL)
		return NULL;
  if ((char *)strchr(buffer, '\n') != NULL)
    *((char *)strchr(buffer, '\n')) = '\0';
  if (buffer[0] == '\0') {
	ptr = malloc(sizeof(char));
	*ptr = '\0';
  } else {
	ptr = malloc((strlen(buffer)+1)*sizeof(char));
	sprintf(ptr, "%s", buffer);
  }
  return ptr;
}

/*
  dcfp_destroy_var_tree()
  free an entire tree of variable structs
*/
static void dcfp_destroy_var_tree(dcfp_var_struct *vars) {
  dcfp_var_struct *ptr=NULL, *tmp=NULL;
  ptr = vars;
  while (ptr != NULL) {
//	varcount++;
	free(ptr->key);
	if (strncmp(ptr->type, "int", 3) == 0) { // free as int
	  if(ptr->value !=NULL) free((int *)(ptr->value));
	} else { // free as char*
	  if(ptr->value !=NULL) free((char *)(ptr->value));
	}
	tmp = ptr;
	ptr = ptr->next;
	free(tmp);
  }
}

/*
  dcfp_destroy_section()
  free an entire section, including subsections
*/
static void dcfp_destroy_section(dcfp_section_struct *sect) {
  dcfp_section_struct *ptr=NULL, *tmp=NULL;
  ptr = sect;
//  sectcount++;
  while (ptr != NULL) {
    free(ptr->name);
	dcfp_destroy_var_tree(ptr->globalvars);
	dcfp_destroy_section(ptr->sections);
	tmp = ptr;
	ptr = ptr->next;
	free(tmp);
  }
}

/*
  is_int()
  check if a string really is an integer
*/
static int is_int (char *str) {
  int isint = 1;
  char *ptr = str;
  
  while (*ptr!='\0') {
	if (48<=*ptr && *ptr<=57) {
	  ptr++;
	  continue;
    } else {
	  isint=0;
	  break;
	}
  }
  return isint;
}


